@@ -5,7 +5,7 @@ use crate::deprecations::Deprecation;
 use crate::params::{accept_args_kwargs, impl_arg_params};
 use crate::pyfunction::PyFunctionOptions;
 use crate::pyfunction::{PyFunctionArgPyO3Attributes, PyFunctionSignature};
-use crate::utils::{self, PythonDoc};
+use crate::utils::{self, get_pyo3_crate, PythonDoc};
 use crate::{deprecations::Deprecations, pyfunction::Argument};
 use proc_macro2::{Span, TokenStream};
 use quote::ToTokens;
@@ -64,17 +64,17 @@ impl<'a> FnArg<'a> {
 
 #[derive(Clone, PartialEq, Debug, Copy, Eq)]
 pub enum MethodTypeAttribute {
-    /// #[new]
+    /// `#[new]`
     New,
-    /// #[classmethod]
+    /// `#[classmethod]`
     ClassMethod,
-    /// #[classattr]
+    /// `#[classattr]`
     ClassAttribute,
-    /// #[staticmethod]
+    /// `#[staticmethod]`
     StaticMethod,
-    /// #[getter]
+    /// `#[getter]`
     Getter,
-    /// #[setter]
+    /// `#[setter]`
     Setter,
 }
 
@@ -106,12 +106,12 @@ impl FnType {
             }
             FnType::FnClass => {
                 quote! {
-                    let _slf = ::pyo3::types::PyType::from_type_ptr(_py, _slf as *mut ::pyo3::ffi::PyTypeObject);
+                    let _slf = _pyo3::types::PyType::from_type_ptr(_py, _slf as *mut _pyo3::ffi::PyTypeObject);
                 }
             }
             FnType::FnModule => {
                 quote! {
-                    let _slf = _py.from_borrowed_ptr::<::pyo3::types::PyModule>(_slf);
+                    let _slf = _py.from_borrowed_ptr::<_pyo3::types::PyModule>(_slf);
                 }
             }
         }
@@ -141,13 +141,13 @@ impl SelfType {
     pub fn receiver(&self, cls: &syn::Type, error_mode: ExtractErrorMode) -> TokenStream {
         let cell = match error_mode {
             ExtractErrorMode::Raise => {
-                quote! { _py.from_borrowed_ptr::<::pyo3::PyAny>(_slf).downcast::<::pyo3::PyCell<#cls>>()? }
+                quote! { _py.from_borrowed_ptr::<_pyo3::PyAny>(_slf).downcast::<_pyo3::PyCell<#cls>>()? }
             }
             ExtractErrorMode::NotImplemented => {
                 quote! {
-                    match _py.from_borrowed_ptr::<::pyo3::PyAny>(_slf).downcast::<::pyo3::PyCell<#cls>>() {
+                    match _py.from_borrowed_ptr::<_pyo3::PyAny>(_slf).downcast::<_pyo3::PyCell<#cls>>() {
                         ::std::result::Result::Ok(cell) => cell,
-                        ::std::result::Result::Err(_) => return ::pyo3::callback::convert(_py, _py.NotImplemented()),
+                        ::std::result::Result::Err(_) => return _pyo3::callback::convert(_py, _py.NotImplemented()),
                     }
                 }
             }
@@ -228,6 +228,8 @@ pub struct FnSpec<'a> {
     pub deprecations: Deprecations,
     pub convention: CallingConvention,
     pub text_signature: Option<TextSignatureAttribute>,
+    pub krate: syn::Path,
+    pub unsafety: Option<syn::Token![unsafe]>,
 }
 
 pub fn get_return_info(output: &syn::ReturnType) -> syn::Type {
@@ -254,12 +256,14 @@ pub fn parse_method_receiver(arg: &syn::FnArg) -> Result<SelfType> {
 impl<'a> FnSpec<'a> {
     /// Parser function signature and function attributes
     pub fn parse(
+        // Signature is mutable to remove the `Python` argument.
         sig: &'a mut syn::Signature,
         meth_attrs: &mut Vec<syn::Attribute>,
         options: PyFunctionOptions,
     ) -> Result<FnSpec<'a>> {
         let PyFunctionOptions {
             text_signature,
+            krate,
             name,
             mut deprecations,
             ..
@@ -278,6 +282,7 @@ impl<'a> FnSpec<'a> {
         let name = &sig.ident;
         let ty = get_return_info(&sig.output);
         let python_name = python_name.as_ref().unwrap_or(name).unraw();
+        let krate = get_pyo3_crate(&krate);
 
         let doc = utils::get_doc(
             meth_attrs,
@@ -311,6 +316,8 @@ impl<'a> FnSpec<'a> {
             doc,
             deprecations,
             text_signature,
+            krate,
+            unsafety: sig.unsafety,
         })
     }
 
@@ -352,15 +359,12 @@ impl<'a> FnSpec<'a> {
             parse_method_receiver(first_arg)
         };
 
-        #[allow(clippy::manual_strip)] // for strip_prefix replacement supporting rust < 1.45
         // strip get_ or set_
         let strip_fn_name = |prefix: &'static str| {
-            let ident = name.unraw().to_string();
-            if ident.starts_with(prefix) {
-                Some(syn::Ident::new(&ident[prefix.len()..], ident.span()))
-            } else {
-                None
-            }
+            name.unraw()
+                .to_string()
+                .strip_prefix(prefix)
+                .map(|stripped| syn::Ident::new(stripped, name.span()))
         };
 
         let (fn_type, skip_first_arg, fixed_convention) = match fn_type_attr {
@@ -416,7 +420,7 @@ impl<'a> FnSpec<'a> {
     }
 
     pub fn default_value(&self, name: &syn::Ident) -> Option<TokenStream> {
-        for s in self.attrs.iter() {
+        for s in &self.attrs {
             match s {
                 Argument::Arg(path, opt) | Argument::Kwarg(path, opt) => {
                     if path.is_ident(name) {
@@ -433,7 +437,7 @@ impl<'a> FnSpec<'a> {
     }
 
     pub fn is_pos_only(&self, name: &syn::Ident) -> bool {
-        for s in self.attrs.iter() {
+        for s in &self.attrs {
             if let Argument::PosOnlyArg(path, _) = s {
                 if path.is_ident(name) {
                     return true;
@@ -444,7 +448,7 @@ impl<'a> FnSpec<'a> {
     }
 
     pub fn is_kw_only(&self, name: &syn::Ident) -> bool {
-        for s in self.attrs.iter() {
+        for s in &self.attrs {
             if let Argument::Kwarg(path, _) = s {
                 if path.is_ident(name) {
                     return true;
@@ -474,89 +478,89 @@ impl<'a> FnSpec<'a> {
             quote!(#func_name)
         };
         let rust_call =
-            quote! { ::pyo3::callback::convert(#py, #rust_name(#self_arg #(#arg_names),*)) };
+            quote! { _pyo3::callback::convert(#py, #rust_name(#self_arg #(#arg_names),*)) };
+        let krate = &self.krate;
         Ok(match self.convention {
             CallingConvention::Noargs => {
                 quote! {
                     unsafe extern "C" fn #ident (
-                        _slf: *mut ::pyo3::ffi::PyObject,
-                        _args: *mut ::pyo3::ffi::PyObject,
-                    ) -> *mut ::pyo3::ffi::PyObject
+                        _slf: *mut #krate::ffi::PyObject,
+                        _args: *mut #krate::ffi::PyObject,
+                    ) -> *mut #krate::ffi::PyObject
                     {
+                        use #krate as _pyo3;
                         #deprecations
-                        ::pyo3::callback::handle_panic(|#py| {
+                        let gil = _pyo3::GILPool::new();
+                        let #py = gil.python();
+                        _pyo3::callback::panic_result_into_callback_output(#py, ::std::panic::catch_unwind(move || -> _pyo3::PyResult<_> {
                             #self_conversion
                             #rust_call
-                        })
+                        }))
                     }
                 }
             }
             CallingConvention::Fastcall => {
-                let arg_convert_and_rust_call = impl_arg_params(self, cls, rust_call, &py, true)?;
+                let arg_convert = impl_arg_params(self, cls, &py, true)?;
                 quote! {
                     unsafe extern "C" fn #ident (
-                        _slf: *mut ::pyo3::ffi::PyObject,
-                        _args: *const *mut ::pyo3::ffi::PyObject,
-                        _nargs: ::pyo3::ffi::Py_ssize_t,
-                        _kwnames: *mut ::pyo3::ffi::PyObject) -> *mut ::pyo3::ffi::PyObject
+                        _slf: *mut #krate::ffi::PyObject,
+                        _args: *const *mut #krate::ffi::PyObject,
+                        _nargs: #krate::ffi::Py_ssize_t,
+                        _kwnames: *mut #krate::ffi::PyObject) -> *mut #krate::ffi::PyObject
                     {
+                        use #krate as _pyo3;
                         #deprecations
-                        ::pyo3::callback::handle_panic(|#py| {
+                        let gil = _pyo3::GILPool::new();
+                        let #py = gil.python();
+                        _pyo3::callback::panic_result_into_callback_output(#py, ::std::panic::catch_unwind(move || -> _pyo3::PyResult<_> {
                             #self_conversion
-                            let _kwnames: ::std::option::Option<&::pyo3::types::PyTuple> = #py.from_borrowed_ptr_or_opt(_kwnames);
-                            // Safety: &PyAny has the same memory layout as `*mut ffi::PyObject`
-                            let _args = _args as *const &::pyo3::PyAny;
-                            let _kwargs = if let ::std::option::Option::Some(kwnames) = _kwnames {
-                                ::std::slice::from_raw_parts(_args.offset(_nargs), kwnames.len())
-                            } else {
-                                &[]
-                            };
-                            let _args = ::std::slice::from_raw_parts(_args, _nargs as usize);
-
-                            #arg_convert_and_rust_call
-                        })
+                            #arg_convert
+                            #rust_call
+                        }))
                     }
                 }
             }
             CallingConvention::Varargs => {
-                let arg_convert_and_rust_call = impl_arg_params(self, cls, rust_call, &py, false)?;
+                let arg_convert = impl_arg_params(self, cls, &py, false)?;
                 quote! {
                     unsafe extern "C" fn #ident (
-                        _slf: *mut ::pyo3::ffi::PyObject,
-                        _args: *mut ::pyo3::ffi::PyObject,
-                        _kwargs: *mut ::pyo3::ffi::PyObject) -> *mut ::pyo3::ffi::PyObject
+                        _slf: *mut #krate::ffi::PyObject,
+                        _args: *mut #krate::ffi::PyObject,
+                        _kwargs: *mut #krate::ffi::PyObject) -> *mut #krate::ffi::PyObject
                     {
+                        use #krate as _pyo3;
                         #deprecations
-                        ::pyo3::callback::handle_panic(|#py| {
+                        let gil = _pyo3::GILPool::new();
+                        let #py = gil.python();
+                        _pyo3::callback::panic_result_into_callback_output(#py, ::std::panic::catch_unwind(move || -> _pyo3::PyResult<_> {
                             #self_conversion
-                            let _args = #py.from_borrowed_ptr::<::pyo3::types::PyTuple>(_args);
-                            let _kwargs: ::std::option::Option<&::pyo3::types::PyDict> = #py.from_borrowed_ptr_or_opt(_kwargs);
-
-                            #arg_convert_and_rust_call
-                        })
+                            #arg_convert
+                            #rust_call
+                        }))
                     }
                 }
             }
             CallingConvention::TpNew => {
                 let rust_call = quote! { #rust_name(#(#arg_names),*) };
-                let arg_convert_and_rust_call = impl_arg_params(self, cls, rust_call, &py, false)?;
+                let arg_convert = impl_arg_params(self, cls, &py, false)?;
                 quote! {
                     unsafe extern "C" fn #ident (
-                        subtype: *mut ::pyo3::ffi::PyTypeObject,
-                        _args: *mut ::pyo3::ffi::PyObject,
-                        _kwargs: *mut ::pyo3::ffi::PyObject) -> *mut ::pyo3::ffi::PyObject
+                        subtype: *mut #krate::ffi::PyTypeObject,
+                        _args: *mut #krate::ffi::PyObject,
+                        _kwargs: *mut #krate::ffi::PyObject) -> *mut #krate::ffi::PyObject
                     {
+                        use #krate as _pyo3;
                         #deprecations
-                        use ::pyo3::callback::IntoPyCallbackOutput;
-                        ::pyo3::callback::handle_panic(|#py| {
-                            let _args = #py.from_borrowed_ptr::<::pyo3::types::PyTuple>(_args);
-                            let _kwargs: ::std::option::Option<&::pyo3::types::PyDict> = #py.from_borrowed_ptr_or_opt(_kwargs);
-
-                            let result = #arg_convert_and_rust_call;
-                            let initializer: ::pyo3::PyClassInitializer::<#cls> = result.convert(#py)?;
+                        use _pyo3::callback::IntoPyCallbackOutput;
+                        let gil = _pyo3::GILPool::new();
+                        let #py = gil.python();
+                        _pyo3::callback::panic_result_into_callback_output(#py, ::std::panic::catch_unwind(move || -> _pyo3::PyResult<_> {
+                            #arg_convert
+                            let result = #rust_call;
+                            let initializer: _pyo3::PyClassInitializer::<#cls> = result.convert(#py)?;
                             let cell = initializer.create_cell_from_subtype(#py, subtype)?;
-                            ::std::result::Result::Ok(cell as *mut ::pyo3::ffi::PyObject)
-                        })
+                            ::std::result::Result::Ok(cell as *mut _pyo3::ffi::PyObject)
+                        }))
                     }
                 }
             }
@@ -570,23 +574,23 @@ impl<'a> FnSpec<'a> {
         let doc = &self.doc;
         match self.convention {
             CallingConvention::Noargs => quote! {
-                ::pyo3::class::methods::PyMethodDef::noargs(
+                _pyo3::impl_::pymethods::PyMethodDef::noargs(
                     #python_name,
-                    ::pyo3::class::methods::PyCFunction(#wrapper),
+                    _pyo3::impl_::pymethods::PyCFunction(#wrapper),
                     #doc,
                 )
             },
             CallingConvention::Fastcall => quote! {
-                ::pyo3::class::methods::PyMethodDef::fastcall_cfunction_with_keywords(
+                _pyo3::impl_::pymethods::PyMethodDef::fastcall_cfunction_with_keywords(
                     #python_name,
-                    ::pyo3::class::methods::PyCFunctionFastWithKeywords(#wrapper),
+                    _pyo3::impl_::pymethods::PyCFunctionFastWithKeywords(#wrapper),
                     #doc,
                 )
             },
             CallingConvention::Varargs => quote! {
-                ::pyo3::class::methods::PyMethodDef::cfunction_with_keywords(
+                _pyo3::impl_::pymethods::PyMethodDef::cfunction_with_keywords(
                     #python_name,
-                    ::pyo3::class::methods::PyCFunctionWithKeywords(#wrapper),
+                    _pyo3::impl_::pymethods::PyCFunctionWithKeywords(#wrapper),
                     #doc,
                 )
             },
