@@ -53,12 +53,11 @@ fn is_kwargs(attrs: &[Argument], name: &syn::Ident) -> bool {
 pub fn impl_arg_params(
     spec: &FnSpec<'_>,
     self_: Option<&syn::Type>,
-    body: TokenStream,
     py: &syn::Ident,
     fastcall: bool,
 ) -> Result<TokenStream> {
     if spec.args.is_empty() {
-        return Ok(body);
+        return Ok(TokenStream::new());
     }
 
     let args_array = syn::Ident::new("output", Span::call_site());
@@ -70,11 +69,11 @@ pub fn impl_arg_params(
         for (i, arg) in spec.args.iter().enumerate() {
             arg_convert.push(impl_arg_param(arg, spec, i, None, &mut 0, py, &args_array)?);
         }
-        return Ok(quote! {{
-            let _args = Some(_args);
+        return Ok(quote! {
+            let _args = #py.from_borrowed_ptr::<_pyo3::types::PyTuple>(_args);
+            let _kwargs: ::std::option::Option<&_pyo3::types::PyDict> = #py.from_borrowed_ptr_or_opt(_kwargs);
             #(#arg_convert)*
-            #body
-        }});
+        });
     };
 
     let mut positional_parameter_names = Vec::new();
@@ -82,7 +81,7 @@ pub fn impl_arg_params(
     let mut required_positional_parameters = 0usize;
     let mut keyword_only_parameters = Vec::new();
 
-    for arg in spec.args.iter() {
+    for arg in &spec.args {
         if arg.py || is_args(&spec.attrs, arg.name) || is_kwargs(&spec.attrs, arg.name) {
             continue;
         }
@@ -93,19 +92,20 @@ pub fn impl_arg_params(
 
         if kwonly {
             keyword_only_parameters.push(quote! {
-                ::pyo3::derive_utils::KeywordOnlyParameterDescription {
+                _pyo3::impl_::extract_argument::KeywordOnlyParameterDescription {
                     name: #name,
                     required: #required,
                 }
             });
         } else {
+            positional_parameter_names.push(name);
+
             if required {
-                required_positional_parameters += 1;
+                required_positional_parameters = positional_parameter_names.len();
             }
             if posonly {
                 positional_only_parameters += 1;
             }
-            positional_parameter_names.push(name);
         }
     }
 
@@ -126,58 +126,61 @@ pub fn impl_arg_params(
     }
 
     let (accept_args, accept_kwargs) = accept_args_kwargs(&spec.attrs);
+    let args_handler = if accept_args {
+        quote! { _pyo3::impl_::extract_argument::TupleVarargs }
+    } else {
+        quote! { _pyo3::impl_::extract_argument::NoVarargs }
+    };
+    let kwargs_handler = if accept_kwargs {
+        quote! { _pyo3::impl_::extract_argument::DictVarkeywords }
+    } else {
+        quote! { _pyo3::impl_::extract_argument::NoVarkeywords }
+    };
 
     let cls_name = if let Some(cls) = self_ {
-        quote! { ::std::option::Option::Some(<#cls as ::pyo3::type_object::PyTypeInfo>::NAME) }
+        quote! { ::std::option::Option::Some(<#cls as _pyo3::type_object::PyTypeInfo>::NAME) }
     } else {
         quote! { ::std::option::Option::None }
     };
     let python_name = &spec.python_name;
 
-    let (args_to_extract, kwargs_to_extract) = if fastcall {
-        // _args is a &[&PyAny], _kwnames is a Option<&PyTuple> containing the
-        // keyword names of the keyword args in _kwargs
-        (
-            // need copied() for &&PyAny -> &PyAny
-            quote! { ::std::iter::Iterator::copied(_args.iter()) },
-            quote! { _kwnames.map(|kwnames| {
-                use ::std::iter::Iterator;
-                kwnames.as_slice().iter().copied().zip(_kwargs.iter().copied())
-            }) },
-        )
+    let extract_expression = if fastcall {
+        quote! {
+            DESCRIPTION.extract_arguments_fastcall::<#args_handler, #kwargs_handler>(
+                #py,
+                _args,
+                _nargs,
+                _kwnames,
+                &mut #args_array
+            )?
+        }
     } else {
-        // _args is a &PyTuple, _kwargs is an Option<&PyDict>
-        (
-            quote! { _args.iter() },
-            quote! { _kwargs.map(|dict| dict.iter()) },
-        )
+        quote! {
+            DESCRIPTION.extract_arguments_tuple_dict::<#args_handler, #kwargs_handler>(
+                #py,
+                _args,
+                _kwargs,
+                &mut #args_array
+            )?
+        }
     };
 
     // create array of arguments, and then parse
-    Ok(quote! {{
-            const DESCRIPTION: ::pyo3::derive_utils::FunctionDescription = ::pyo3::derive_utils::FunctionDescription {
+    Ok(quote! {
+            const DESCRIPTION: _pyo3::impl_::extract_argument::FunctionDescription = _pyo3::impl_::extract_argument::FunctionDescription {
                 cls_name: #cls_name,
                 func_name: stringify!(#python_name),
                 positional_parameter_names: &[#(#positional_parameter_names),*],
                 positional_only_parameters: #positional_only_parameters,
                 required_positional_parameters: #required_positional_parameters,
                 keyword_only_parameters: &[#(#keyword_only_parameters),*],
-                accept_varargs: #accept_args,
-                accept_varkeywords: #accept_kwargs,
             };
 
             let mut #args_array = [::std::option::Option::None; #num_params];
-            let (_args, _kwargs) = DESCRIPTION.extract_arguments(
-                #py,
-                #args_to_extract,
-                #kwargs_to_extract,
-                &mut #args_array
-            )?;
+            let (_args, _kwargs) = #extract_expression;
 
             #(#param_conversion)*
-
-            #body
-    }})
+    })
 }
 
 /// Re option_pos: The option slice doesn't contain the py: Python argument, so the argument
@@ -205,9 +208,7 @@ fn impl_arg_param(
 
     let ty = arg.ty;
     let name = arg.name;
-    let transform_error = quote! {
-        |e| ::pyo3::derive_utils::argument_extraction_error(#py, stringify!(#name), e)
-    };
+    let name_str = name.to_string();
 
     if is_args(&spec.attrs, name) {
         ensure_spanned!(
@@ -215,7 +216,7 @@ fn impl_arg_param(
             arg.name.span() => "args cannot be optional"
         );
         return Ok(quote_arg_span! {
-            let #arg_name = _args.unwrap().extract().map_err(#transform_error)?;
+            let #arg_name = _pyo3::impl_::extract_argument::extract_argument(_args, #name_str)?;
         });
     } else if is_kwargs(&spec.attrs, name) {
         ensure_spanned!(
@@ -223,44 +224,64 @@ fn impl_arg_param(
             arg.name.span() => "kwargs must be Option<_>"
         );
         return Ok(quote_arg_span! {
-            let #arg_name = _kwargs.map(|kwargs| kwargs.extract())
-                .transpose()
-                .map_err(#transform_error)?;
+            let #arg_name = _pyo3::impl_::extract_argument::extract_optional_argument(_kwargs.map(|kwargs| kwargs.as_ref()), #name_str)?;
         });
     }
 
     let arg_value = quote_arg_span!(#args_array[#option_pos]);
     *option_pos += 1;
 
-    let extract = if let Some(FromPyWithAttribute(expr_path)) = &arg.attrs.from_py_with {
-        quote_arg_span! { #expr_path(_obj).map_err(#transform_error) }
+    let arg_value_or_default = if let Some(FromPyWithAttribute(expr_path)) = &arg.attrs.from_py_with
+    {
+        match (spec.default_value(name), arg.optional.is_some()) {
+            (Some(default), true) if default.to_string() != "None" => {
+                quote_arg_span! {
+                    _pyo3::impl_::extract_argument::from_py_with_with_default(#arg_value, #name_str, #expr_path, || Some(#default))
+                }
+            }
+            (Some(default), _) => {
+                quote_arg_span! {
+                    _pyo3::impl_::extract_argument::from_py_with_with_default(#arg_value, #name_str, #expr_path, || #default)
+                }
+            }
+            (None, true) => {
+                quote_arg_span! {
+                    _pyo3::impl_::extract_argument::from_py_with_with_default(#arg_value, #name_str, #expr_path, || Some(None))
+                }
+            }
+            (None, false) => {
+                quote_arg_span! {
+                    _pyo3::impl_::extract_argument::from_py_with(
+                        _pyo3::impl_::extract_argument::unwrap_required_argument(#arg_value),
+                        #name_str,
+                        #expr_path,
+                    )
+                }
+            }
+        }
     } else {
-        quote_arg_span! { _obj.extract().map_err(#transform_error) }
-    };
-
-    let arg_value_or_default = match (spec.default_value(name), arg.optional.is_some()) {
-        (Some(default), true) if default.to_string() != "None" => {
-            quote_arg_span! {
-                #arg_value.map_or_else(|| ::std::result::Result::Ok(::std::option::Option::Some(#default)),
-                                       |_obj| #extract)?
+        match (spec.default_value(name), arg.optional.is_some()) {
+            (Some(default), true) if default.to_string() != "None" => {
+                quote_arg_span! {
+                    _pyo3::impl_::extract_argument::extract_argument_with_default(#arg_value, #name_str, || Some(#default))
+                }
             }
-        }
-        (Some(default), _) => {
-            quote_arg_span! {
-                #arg_value.map_or_else(|| ::std::result::Result::Ok(#default), |_obj| #extract)?
+            (Some(default), _) => {
+                quote_arg_span! {
+                    _pyo3::impl_::extract_argument::extract_argument_with_default(#arg_value, #name_str, || #default)
+                }
             }
-        }
-        (None, true) => {
-            quote_arg_span! {
-                #arg_value.map_or(::std::result::Result::Ok(::std::option::Option::None),
-                                  |_obj| #extract)?
+            (None, true) => {
+                quote_arg_span! {
+                    _pyo3::impl_::extract_argument::extract_optional_argument(#arg_value, #name_str)
+                }
             }
-        }
-        (None, false) => {
-            quote_arg_span! {
-                {
-                    let _obj = #arg_value.expect("Failed to extract required method argument");
-                    #extract?
+            (None, false) => {
+                quote_arg_span! {
+                    _pyo3::impl_::extract_argument::extract_argument(
+                        _pyo3::impl_::extract_argument::unwrap_required_argument(#arg_value),
+                        #name_str
+                    )
                 }
             }
         }
@@ -275,7 +296,7 @@ fn impl_arg_param(
         let (target_ty, borrow_tmp) = if arg.optional.is_some() {
             // Get Option<&T> from Option<PyRef<T>>
             (
-                quote_arg_span! { ::std::option::Option<<#tref as ::pyo3::derive_utils::ExtractExt<'_>>::Target> },
+                quote_arg_span! { ::std::option::Option<<#tref as _pyo3::derive_utils::ExtractExt<'_>>::Target> },
                 if mut_.is_some() {
                     quote_arg_span! { _tmp.as_deref_mut() }
                 } else {
@@ -285,18 +306,19 @@ fn impl_arg_param(
         } else {
             // Get &T from PyRef<T>
             (
-                quote_arg_span! { <#tref as ::pyo3::derive_utils::ExtractExt<'_>>::Target },
+                quote_arg_span! { <#tref as _pyo3::derive_utils::ExtractExt<'_>>::Target },
                 quote_arg_span! { &#mut_ *_tmp },
             )
         };
 
         Ok(quote_arg_span! {
-            let #mut_ _tmp: #target_ty = #arg_value_or_default;
+            let #mut_ _tmp: #target_ty = #arg_value_or_default?;
+            #[allow(clippy::needless_option_as_deref)]
             let #arg_name = #borrow_tmp;
         })
     } else {
         Ok(quote_arg_span! {
-            let #arg_name = #arg_value_or_default;
+            let #arg_name = #arg_value_or_default?;
         })
     };
 }
