@@ -2,19 +2,18 @@
 
 use crate::{
     attributes::{
-        self, get_deprecated_name_attribute, get_deprecated_text_signature_attribute,
-        get_pyo3_options, take_attributes, FromPyWithAttribute, NameAttribute,
-        TextSignatureAttribute,
+        self, get_pyo3_options, take_attributes, take_pyo3_options, CrateAttribute,
+        FromPyWithAttribute, NameAttribute, TextSignatureAttribute,
     },
     deprecations::Deprecations,
     method::{self, CallingConvention, FnArg},
     pymethod::check_generic,
-    utils::{self, ensure_not_async_fn},
+    utils::{self, ensure_not_async_fn, get_pyo3_crate},
 };
 use proc_macro2::{Span, TokenStream};
 use quote::{format_ident, quote};
 use syn::punctuated::Punctuated;
-use syn::{ext::IdentExt, spanned::Spanned, Ident, NestedMeta, Path, Result};
+use syn::{ext::IdentExt, spanned::Spanned, NestedMeta, Path, Result};
 use syn::{
     parse::{Parse, ParseBuffer, ParseStream},
     token::Comma,
@@ -22,9 +21,11 @@ use syn::{
 
 #[derive(Debug, Clone, PartialEq)]
 pub enum Argument {
+    PosOnlyArgsSeparator,
     VarArgsSeparator,
     VarArgs(syn::Path),
     KeywordArgs(syn::Path),
+    PosOnlyArg(syn::Path, Option<String>),
     Arg(syn::Path, Option<String>),
     Kwarg(syn::Path, Option<String>),
 }
@@ -34,6 +35,7 @@ pub enum Argument {
 pub struct PyFunctionSignature {
     pub arguments: Vec<Argument>,
     has_kw: bool,
+    has_posonly_args: bool,
     has_varargs: bool,
     has_kwargs: bool,
 }
@@ -126,7 +128,22 @@ impl PyFunctionSignature {
                 self.arguments.push(Argument::VarArgsSeparator);
                 Ok(())
             }
-            _ => bail_spanned!(item.span() => "expected \"*\""),
+            syn::Lit::Str(lits) if lits.value() == "/" => {
+                // "/"
+                self.posonly_arg_is_ok(item)?;
+                self.has_posonly_args = true;
+                // any arguments _before_ this become positional-only
+                self.arguments.iter_mut().for_each(|a| {
+                    if let Argument::Arg(path, name) = a {
+                        *a = Argument::PosOnlyArg(path.clone(), name.clone());
+                    } else {
+                        unreachable!();
+                    }
+                });
+                self.arguments.push(Argument::PosOnlyArgsSeparator);
+                Ok(())
+            }
+            _ => bail_spanned!(item.span() => "expected \"/\" or \"*\""),
         }
     }
 
@@ -143,6 +160,14 @@ impl PyFunctionSignature {
         Ok(())
     }
 
+    fn posonly_arg_is_ok(&self, item: &NestedMeta) -> syn::Result<()> {
+        ensure_spanned!(
+            !(self.has_posonly_args || self.has_kwargs || self.has_varargs),
+            item.span() => "/ is not allowed after /, varargs(*), or kwargs(**)"
+        );
+        Ok(())
+    }
+
     fn vararg_is_ok(&self, item: &NestedMeta) -> syn::Result<()> {
         ensure_spanned!(
             !(self.has_kwargs || self.has_varargs),
@@ -214,17 +239,12 @@ pub struct PyFunctionOptions {
     pub signature: Option<PyFunctionSignature>,
     pub text_signature: Option<TextSignatureAttribute>,
     pub deprecations: Deprecations,
+    pub krate: Option<CrateAttribute>,
 }
 
 impl Parse for PyFunctionOptions {
     fn parse(input: ParseStream) -> Result<Self> {
-        let mut options = PyFunctionOptions {
-            pass_module: None,
-            name: None,
-            signature: None,
-            text_signature: None,
-            deprecations: Deprecations::new(),
-        };
+        let mut options = PyFunctionOptions::default();
 
         while !input.is_empty() {
             let lookahead = input.lookahead1();
@@ -237,6 +257,9 @@ impl Parse for PyFunctionOptions {
                 if !input.is_empty() {
                     let _: Comma = input.parse()?;
                 }
+            } else if lookahead.peek(syn::Token![crate]) {
+                // TODO needs duplicate check?
+                options.krate = Some(input.parse()?);
             } else {
                 // If not recognised attribute, this is "legacy" pyfunction syntax #[pyfunction(a, b)]
                 //
@@ -255,6 +278,7 @@ pub enum PyFunctionOption {
     PassModule(attributes::kw::pass_module),
     Signature(PyFunctionSignature),
     TextSignature(TextSignatureAttribute),
+    Crate(CrateAttribute),
 }
 
 impl Parse for PyFunctionOption {
@@ -268,6 +292,8 @@ impl Parse for PyFunctionOption {
             input.parse().map(PyFunctionOption::Signature)
         } else if lookahead.peek(attributes::kw::text_signature) {
             input.parse().map(PyFunctionOption::TextSignature)
+        } else if lookahead.peek(syn::Token![crate]) {
+            input.parse().map(PyFunctionOption::Crate)
         } else {
             Err(lookahead.error())
         }
@@ -277,34 +303,10 @@ impl Parse for PyFunctionOption {
 impl PyFunctionOptions {
     pub fn from_attrs(attrs: &mut Vec<syn::Attribute>) -> syn::Result<Self> {
         let mut options = PyFunctionOptions::default();
-        options.take_pyo3_options(attrs)?;
+        options.add_attributes(take_pyo3_options(attrs)?)?;
         Ok(options)
     }
 
-    pub fn take_pyo3_options(&mut self, attrs: &mut Vec<syn::Attribute>) -> syn::Result<()> {
-        take_attributes(attrs, |attr| {
-            if let Some(pyo3_attributes) = get_pyo3_options(attr)? {
-                self.add_attributes(pyo3_attributes)?;
-                Ok(true)
-            } else if let Some(name) = get_deprecated_name_attribute(attr, &mut self.deprecations)?
-            {
-                self.set_name(name)?;
-                Ok(true)
-            } else if let Some(text_signature) =
-                get_deprecated_text_signature_attribute(attr, &mut self.deprecations)?
-            {
-                self.add_attributes(std::iter::once(PyFunctionOption::TextSignature(
-                    text_signature,
-                )))?;
-                Ok(true)
-            } else {
-                Ok(false)
-            }
-        })?;
-
-        Ok(())
-    }
-
     pub fn add_attributes(
         &mut self,
         attrs: impl IntoIterator<Item = PyFunctionOption>,
@@ -334,6 +336,13 @@ impl PyFunctionOptions {
                     );
                     self.text_signature = Some(text_signature);
                 }
+                PyFunctionOption::Crate(path) => {
+                    ensure_spanned!(
+                        self.krate.is_none(),
+                        path.0.span() => "`crate` may only be specified once"
+                    );
+                    self.krate = Some(path);
+                }
             }
         }
         Ok(())
@@ -353,14 +362,8 @@ pub fn build_py_function(
     ast: &mut syn::ItemFn,
     mut options: PyFunctionOptions,
 ) -> syn::Result<TokenStream> {
-    options.take_pyo3_options(&mut ast.attrs)?;
-    Ok(impl_wrap_pyfunction(ast, options)?.1)
-}
-
-/// Coordinates the naming of a the add-function-to-python-module function
-fn function_wrapper_ident(name: &Ident) -> Ident {
-    // Make sure this ident matches the one of wrap_pyfunction
-    format_ident!("__pyo3_get_function_{}", name)
+    options.add_attributes(take_pyo3_options(&mut ast.attrs)?)?;
+    impl_wrap_pyfunction(ast, options)
 }
 
 /// Generates python wrapper over a function that allows adding it to a python module as a python
@@ -368,7 +371,7 @@ fn function_wrapper_ident(name: &Ident) -> Ident {
 pub fn impl_wrap_pyfunction(
     func: &mut syn::ItemFn,
     options: PyFunctionOptions,
-) -> syn::Result<(Ident, TokenStream)> {
+) -> syn::Result<TokenStream> {
     check_generic(&func.sig)?;
     ensure_not_async_fn(&func.sig)?;
 
@@ -406,9 +409,9 @@ pub fn impl_wrap_pyfunction(
             .text_signature
             .as_ref()
             .map(|attr| (&python_name, attr)),
-    )?;
+    );
 
-    let function_wrapper_ident = function_wrapper_ident(&func.sig.ident);
+    let krate = get_pyo3_crate(&options.krate);
 
     let spec = method::FnSpec {
         tp: if options.pass_module.is_some() {
@@ -424,21 +427,45 @@ pub fn impl_wrap_pyfunction(
         output: ty,
         doc,
         deprecations: options.deprecations,
+        text_signature: options.text_signature,
+        krate: krate.clone(),
+        unsafety: func.sig.unsafety,
     };
 
-    let wrapper_ident = format_ident!("__pyo3_raw_{}", spec.name);
+    let vis = &func.vis;
+    let name = &func.sig.ident;
+
+    let wrapper_ident = format_ident!("__pyfunction_{}", spec.name);
     let wrapper = spec.get_wrapper_function(&wrapper_ident, None)?;
     let methoddef = spec.get_methoddef(wrapper_ident);
 
     let wrapped_pyfunction = quote! {
         #wrapper
-        pub(crate) fn #function_wrapper_ident<'a>(
-            args: impl Into<pyo3::derive_utils::PyFunctionArguments<'a>>
-        ) -> pyo3::PyResult<&'a pyo3::types::PyCFunction> {
-            pyo3::types::PyCFunction::internal_new(#methoddef, args.into())
+
+        // Create a module with the same name as the `#[pyfunction]` - this way `use <the function>`
+        // will actually bring both the module and the function into scope.
+        #[doc(hidden)]
+        #vis mod #name {
+            use #krate as _pyo3;
+            pub(crate) struct PyO3Def;
+
+            // Exported for `wrap_pyfunction!`
+            pub use _pyo3::impl_::pyfunction::wrap_pyfunction as wrap;
+            pub const DEF: _pyo3::PyMethodDef = <PyO3Def as _pyo3::impl_::pyfunction::PyFunctionDef>::DEF;
         }
+
+        // Generate the definition inside an anonymous function in the same scope as the original function -
+        // this avoids complications around the fact that the generated module has a different scope
+        // (and `super` doesn't always refer to the outer scope, e.g. if the `#[pyfunction] is
+        // inside a function body)
+        const _: () = {
+            use #krate as _pyo3;
+            impl _pyo3::impl_::pyfunction::PyFunctionDef for #name::PyO3Def {
+                const DEF: _pyo3::PyMethodDef = #methoddef;
+            }
+        };
     };
-    Ok((function_wrapper_ident, wrapped_pyfunction))
+    Ok(wrapped_pyfunction)
 }
 
 fn type_is_pymodule(ty: &syn::Type) -> bool {
