@@ -3,21 +3,72 @@
 use std::collections::HashSet;
 
 use crate::{
+    attributes::{take_pyo3_options, CrateAttribute},
     konst::{ConstAttributes, ConstSpec},
     pyfunction::PyFunctionOptions,
     pymethod::{self, is_proto_method},
+    utils::get_pyo3_crate,
 };
 use proc_macro2::TokenStream;
 use pymethod::GeneratedPyMethod;
 use quote::quote;
-use syn::spanned::Spanned;
+use syn::{
+    parse::{Parse, ParseStream},
+    spanned::Spanned,
+    Result,
+};
 
 /// The mechanism used to collect `#[pymethods]` into the type object
+#[derive(Copy, Clone)]
 pub enum PyClassMethodsType {
     Specialization,
     Inventory,
 }
 
+enum PyImplPyO3Option {
+    Crate(CrateAttribute),
+}
+
+impl Parse for PyImplPyO3Option {
+    fn parse(input: ParseStream) -> Result<Self> {
+        let lookahead = input.lookahead1();
+        if lookahead.peek(syn::Token![crate]) {
+            input.parse().map(PyImplPyO3Option::Crate)
+        } else {
+            Err(lookahead.error())
+        }
+    }
+}
+
+#[derive(Default)]
+pub struct PyImplOptions {
+    krate: Option<CrateAttribute>,
+}
+
+impl PyImplOptions {
+    pub fn from_attrs(attrs: &mut Vec<syn::Attribute>) -> Result<Self> {
+        let mut options: PyImplOptions = Default::default();
+
+        for option in take_pyo3_options(attrs)? {
+            match option {
+                PyImplPyO3Option::Crate(path) => options.set_crate(path)?,
+            }
+        }
+
+        Ok(options)
+    }
+
+    fn set_crate(&mut self, path: CrateAttribute) -> Result<()> {
+        ensure_spanned!(
+            self.krate.is_none(),
+            path.0.span() => "`crate` may only be specified once"
+        );
+
+        self.krate = Some(path);
+        Ok(())
+    }
+}
+
 pub fn build_py_methods(
     ast: &mut syn::ItemImpl,
     methods_type: PyClassMethodsType,
@@ -30,7 +81,8 @@ pub fn build_py_methods(
             "#[pymethods] cannot be used with lifetime parameters or generics"
         );
     } else {
-        impl_methods(&ast.self_ty, &mut ast.items, methods_type)
+        let options = PyImplOptions::from_attrs(&mut ast.attrs)?;
+        impl_methods(&ast.self_ty, &mut ast.items, methods_type, options)
     }
 }
 
@@ -38,6 +90,7 @@ pub fn impl_methods(
     ty: &syn::Type,
     impls: &mut Vec<syn::ImplItem>,
     methods_type: PyClassMethodsType,
+    options: PyImplOptions,
 ) -> syn::Result<TokenStream> {
     let mut trait_impls = Vec::new();
     let mut proto_impls = Vec::new();
@@ -48,16 +101,13 @@ pub fn impl_methods(
     for iimpl in impls.iter_mut() {
         match iimpl {
             syn::ImplItem::Method(meth) => {
-                let options = PyFunctionOptions::from_attrs(&mut meth.attrs)?;
-                match pymethod::gen_py_method(ty, &mut meth.sig, &mut meth.attrs, options)? {
+                let mut fun_options = PyFunctionOptions::from_attrs(&mut meth.attrs)?;
+                fun_options.krate = fun_options.krate.or_else(|| options.krate.clone());
+                match pymethod::gen_py_method(ty, &mut meth.sig, &mut meth.attrs, fun_options)? {
                     GeneratedPyMethod::Method(token_stream) => {
                         let attrs = get_cfg_attributes(&meth.attrs);
                         methods.push(quote!(#(#attrs)* #token_stream));
                     }
-                    GeneratedPyMethod::TraitImpl(token_stream) => {
-                        let attrs = get_cfg_attributes(&meth.attrs);
-                        trait_impls.push(quote!(#(#attrs)* #token_stream));
-                    }
                     GeneratedPyMethod::SlotTraitImpl(method_name, token_stream) => {
                         implemented_proto_fragments.insert(method_name);
                         let attrs = get_cfg_attributes(&meth.attrs);
@@ -94,42 +144,36 @@ pub fn impl_methods(
 
     add_shared_proto_slots(ty, &mut proto_impls, implemented_proto_fragments);
 
-    Ok(match methods_type {
-        PyClassMethodsType::Specialization => {
-            let methods_registration = impl_py_methods(ty, methods);
-            let protos_registration = impl_protos(ty, proto_impls);
+    let krate = get_pyo3_crate(&options.krate);
 
-            quote! {
-                #(#trait_impls)*
+    let items = match methods_type {
+        PyClassMethodsType::Specialization => impl_py_methods(ty, methods, proto_impls),
+        PyClassMethodsType::Inventory => submit_methods_inventory(ty, methods, proto_impls),
+    };
 
-                #protos_registration
+    Ok(quote! {
+        const _: () = {
+            use #krate as _pyo3;
 
-                #methods_registration
-            }
-        }
-        PyClassMethodsType::Inventory => {
-            let inventory = submit_methods_inventory(ty, methods, proto_impls);
-            quote! {
-                #(#trait_impls)*
+            #(#trait_impls)*
 
-                #inventory
-            }
-        }
+            #items
+        };
     })
 }
 
-fn gen_py_const(cls: &syn::Type, spec: &ConstSpec) -> TokenStream {
+pub fn gen_py_const(cls: &syn::Type, spec: &ConstSpec) -> TokenStream {
     let member = &spec.rust_ident;
     let deprecations = &spec.attributes.deprecations;
     let python_name = &spec.null_terminated_python_name();
     quote! {
-        ::pyo3::class::PyMethodDefType::ClassAttribute({
-            ::pyo3::class::PyClassAttributeDef::new(
+        _pyo3::class::PyMethodDefType::ClassAttribute({
+            _pyo3::class::PyClassAttributeDef::new(
                 #python_name,
-                ::pyo3::class::methods::PyClassAttributeFactory({
-                    fn __wrap(py: ::pyo3::Python<'_>) -> ::pyo3::PyObject {
+                _pyo3::impl_::pymethods::PyClassAttributeFactory({
+                    fn __wrap(py: _pyo3::Python<'_>) -> _pyo3::PyObject {
                         #deprecations
-                        ::pyo3::IntoPy::into_py(#cls::#member, py)
+                        _pyo3::IntoPy::into_py(#cls::#member, py)
                     }
                     __wrap
                 })
@@ -138,14 +182,46 @@ fn gen_py_const(cls: &syn::Type, spec: &ConstSpec) -> TokenStream {
     }
 }
 
-fn impl_py_methods(ty: &syn::Type, methods: Vec<TokenStream>) -> TokenStream {
+pub fn gen_default_items<'a>(
+    cls: &syn::Ident,
+    method_defs: &'a mut [syn::ImplItemMethod],
+) -> impl Iterator<Item = TokenStream> + 'a {
+    // This function uses a lot of `unwrap()`; since method_defs are provided by us, they should
+    // all succeed.
+    let ty: syn::Type = syn::parse_quote!(#cls);
+
+    method_defs.iter_mut().map(move |meth| {
+        let options = PyFunctionOptions::from_attrs(&mut meth.attrs).unwrap();
+        match pymethod::gen_py_method(&ty, &mut meth.sig, &mut meth.attrs, options).unwrap() {
+            GeneratedPyMethod::Proto(token_stream) => {
+                let attrs = get_cfg_attributes(&meth.attrs);
+                quote!(#(#attrs)* #token_stream)
+            }
+            GeneratedPyMethod::SlotTraitImpl(..) => {
+                panic!("SlotFragment methods cannot have default implementation!")
+            }
+            GeneratedPyMethod::Method(_) => {
+                panic!("Only protocol methods can have default implementation!")
+            }
+        }
+    })
+}
+
+fn impl_py_methods(
+    ty: &syn::Type,
+    methods: Vec<TokenStream>,
+    proto_impls: Vec<TokenStream>,
+) -> TokenStream {
     quote! {
-        impl ::pyo3::class::impl_::PyMethods<#ty>
-            for ::pyo3::class::impl_::PyClassImplCollector<#ty>
+        impl _pyo3::impl_::pyclass::PyMethods<#ty>
+            for _pyo3::impl_::pyclass::PyClassImplCollector<#ty>
         {
-            fn py_methods(self) -> &'static [::pyo3::class::methods::PyMethodDefType] {
-                static METHODS: &[::pyo3::class::methods::PyMethodDefType] = &[#(#methods),*];
-                METHODS
+            fn py_methods(self) -> &'static _pyo3::impl_::pyclass::PyClassItems {
+                static ITEMS: _pyo3::impl_::pyclass::PyClassItems = _pyo3::impl_::pyclass::PyClassItems {
+                    methods: &[#(#methods),*],
+                    slots: &[#(#proto_impls),*]
+                };
+                &ITEMS
             }
         }
     }
@@ -161,11 +237,16 @@ fn add_shared_proto_slots(
             let first_implemented = implemented_proto_fragments.remove($first);
             let second_implemented = implemented_proto_fragments.remove($second);
             if first_implemented || second_implemented {
-                proto_impls.push(quote! { ::pyo3::$slot!(#ty) })
+                proto_impls.push(quote! { _pyo3::impl_::pyclass::$slot!(#ty) })
             }
         }};
     }
 
+    try_add_shared_slot!(
+        "__getattribute__",
+        "__getattr__",
+        generate_pyclass_getattro_slot
+    );
     try_add_shared_slot!("__setattr__", "__delattr__", generate_pyclass_setattr_slot);
     try_add_shared_slot!("__set__", "__delete__", generate_pyclass_setdescr_slot);
     try_add_shared_slot!("__setitem__", "__delitem__", generate_pyclass_setitem_slot);
@@ -188,32 +269,20 @@ fn add_shared_proto_slots(
     );
     try_add_shared_slot!("__pow__", "__rpow__", generate_pyclass_pow_slot);
 
+    // if this assertion trips, a slot fragment has been implemented which has not been added in the
+    // list above
     assert!(implemented_proto_fragments.is_empty());
 }
 
-fn impl_protos(ty: &syn::Type, proto_impls: Vec<TokenStream>) -> TokenStream {
-    quote! {
-        impl ::pyo3::class::impl_::PyMethodsProtocolSlots<#ty>
-            for ::pyo3::class::impl_::PyClassImplCollector<#ty>
-        {
-            fn methods_protocol_slots(self) -> &'static [::pyo3::ffi::PyType_Slot] {
-                &[#(#proto_impls),*]
-            }
-        }
-    }
-}
-
 fn submit_methods_inventory(
     ty: &syn::Type,
     methods: Vec<TokenStream>,
     proto_impls: Vec<TokenStream>,
 ) -> TokenStream {
     quote! {
-        ::pyo3::inventory::submit! {
-            #![crate = ::pyo3] {
-                type Inventory = <#ty as ::pyo3::class::impl_::HasMethodsInventory>::Methods;
-                <Inventory as ::pyo3::class::impl_::PyMethodsInventory>::new(::std::vec![#(#methods),*], ::std::vec![#(#proto_impls),*])
-            }
+        _pyo3::inventory::submit! {
+            type Inventory = <#ty as _pyo3::impl_::pyclass::PyClassImpl>::Inventory;
+            Inventory::new(_pyo3::impl_::pyclass::PyClassItems { methods: &[#(#methods),*], slots: &[#(#proto_impls),*] })
         }
     }
 }
