@@ -18,7 +18,6 @@ pub fn build_py_proto(ast: &mut syn::ItemImpl) -> syn::Result<TokenStream> {
             Some(segment) if segment.ident == "PyAsyncProtocol" => &defs::ASYNC,
             Some(segment) if segment.ident == "PyMappingProtocol" => &defs::MAPPING,
             Some(segment) if segment.ident == "PyIterProtocol" => &defs::ITER,
-            Some(segment) if segment.ident == "PyContextProtocol" => &defs::CONTEXT,
             Some(segment) if segment.ident == "PySequenceProtocol" => &defs::SEQ,
             Some(segment) if segment.ident == "PyNumberProtocol" => &defs::NUM,
             Some(segment) if segment.ident == "PyDescrProtocol" => &defs::DESCR,
@@ -68,7 +67,7 @@ fn impl_proto_impl(
 
                 let flags = if m.can_coexist {
                     // We need METH_COEXIST here to prevent __add__  from overriding __radd__
-                    Some(quote!(::pyo3::ffi::METH_COEXIST))
+                    Some(quote!(_pyo3::ffi::METH_COEXIST))
                 } else {
                     None
                 };
@@ -85,76 +84,30 @@ fn impl_proto_impl(
             }
         }
     }
-    let normal_methods = impl_normal_methods(py_methods, ty, proto);
-    let protocol_methods = impl_proto_methods(method_names, ty, proto);
+    let items = impl_proto_items(method_names, py_methods, ty, proto);
+
     Ok(quote! {
-        #trait_impls
-        #normal_methods
-        #protocol_methods
+        const _: () = {
+            use ::pyo3 as _pyo3; // pyproto doesn't support specifying #[pyo3(crate)]
+            #trait_impls
+            #items
+        };
     })
 }
 
-fn impl_normal_methods(
+fn impl_proto_items(
+    method_names: HashSet<String>,
     py_methods: Vec<TokenStream>,
     ty: &syn::Type,
     proto: &defs::Proto,
 ) -> TokenStream {
-    if py_methods.is_empty() {
-        return TokenStream::default();
-    }
-
-    let methods_trait = proto.methods_trait();
-    let methods_trait_methods = proto.methods_trait_methods();
-    quote! {
-        impl ::pyo3::class::impl_::#methods_trait<#ty>
-            for ::pyo3::class::impl_::PyClassImplCollector<#ty>
-        {
-            fn #methods_trait_methods(self) -> &'static [::pyo3::class::methods::PyMethodDefType] {
-                static METHODS: &[::pyo3::class::methods::PyMethodDefType] =
-                    &[#(#py_methods),*];
-                METHODS
-            }
-        }
-    }
-}
-
-fn impl_proto_methods(
-    method_names: HashSet<String>,
-    ty: &syn::Type,
-    proto: &defs::Proto,
-) -> TokenStream {
-    if method_names.is_empty() {
+    if method_names.is_empty() && py_methods.is_empty() {
         return TokenStream::default();
     }
 
     let module = proto.module();
-    let slots_trait = proto.slots_trait();
-    let slots_trait_slots = proto.slots_trait_slots();
-
-    let mut maybe_buffer_methods = None;
-
-    let build_config = pyo3_build_config::get();
-    const PY39: pyo3_build_config::PythonVersion =
-        pyo3_build_config::PythonVersion { major: 3, minor: 9 };
-
-    if build_config.version <= PY39 && proto.name == "Buffer" {
-        maybe_buffer_methods = Some(quote! {
-            impl ::pyo3::class::impl_::PyBufferProtocolProcs<#ty>
-                for ::pyo3::class::impl_::PyClassImplCollector<#ty>
-            {
-                fn buffer_procs(
-                    self
-                ) -> ::std::option::Option<&'static ::pyo3::class::impl_::PyBufferProcs> {
-                    static PROCS: ::pyo3::class::impl_::PyBufferProcs
-                        = ::pyo3::class::impl_::PyBufferProcs {
-                            bf_getbuffer: ::std::option::Option::Some(::pyo3::class::buffer::getbuffer::<#ty>),
-                            bf_releasebuffer: ::std::option::Option::Some(::pyo3::class::buffer::releasebuffer::<#ty>),
-                        };
-                    ::std::option::Option::Some(&PROCS)
-                }
-            }
-        });
-    }
+    let items_trait = proto.items_trait();
+    let items_trait_items = proto.items_trait_items();
 
     let mut tokens = proto
         .slot_defs(method_names)
@@ -162,8 +115,8 @@ fn impl_proto_methods(
             let slot = syn::Ident::new(def.slot, Span::call_site());
             let slot_impl = syn::Ident::new(def.slot_impl, Span::call_site());
             quote! {{
-                ::pyo3::ffi::PyType_Slot {
-                    slot: ::pyo3::ffi::#slot,
+                _pyo3::ffi::PyType_Slot {
+                    slot: _pyo3::ffi::#slot,
                     pfunc: #module::#slot_impl::<#ty> as _
                 }
             }}
@@ -175,13 +128,15 @@ fn impl_proto_methods(
     }
 
     quote! {
-        #maybe_buffer_methods
-
-        impl ::pyo3::class::impl_::#slots_trait<#ty>
-            for ::pyo3::class::impl_::PyClassImplCollector<#ty>
+        impl _pyo3::impl_::pyclass::#items_trait<#ty>
+            for _pyo3::impl_::pyclass::PyClassImplCollector<#ty>
         {
-            fn #slots_trait_slots(self) -> &'static [::pyo3::ffi::PyType_Slot] {
-                &[#(#tokens),*]
+            fn #items_trait_items(self) -> &'static _pyo3::impl_::pyclass::PyClassItems {
+                static ITEMS: _pyo3::impl_::pyclass::PyClassItems = _pyo3::impl_::pyclass::PyClassItems {
+                    methods: &[#(#py_methods),*],
+                    slots: &[#(#tokens),*]
+                };
+                &ITEMS
             }
         }
     }
